// Copyright (c) 2023 University of Florida
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

// Greg Stitt
// University of Florida

// Module Name:  afu.sv
// Project:      bit_diff
// Description: This AFU provides a memory-mapped bit_diff calculator from
//              my SV tutorial. It intentionally has several errors that you
//              must debug using Intel ASE.

//===================================================================
// Interface Description
// clk  : Clock input
// rst  : Reset input (active high)
// mmio : Memory-mapped I/O interface. See mmio_if.vh.
// dma  : DMA interface. See dma_if.vh.
//===================================================================

`include "cci_mpf_if.vh"

module afu 
  (
   input clk,
   input rst,
         mmio_if.user mmio,
         dma_if.peripheral dma
   );

   localparam int INPUT_WIDTH = 64;
   localparam int OUTPUT_WIDTH = $clog2(2*INPUT_WIDTH+1);
   
   logic          go;
   logic          done;
   logic [INPUT_WIDTH-1:0] data;
   logic [OUTPUT_WIDTH-1:0] result;
   
   memory_map
     #(
       .INPUT_WIDTH(INPUT_WIDTH),
       .OUTPUT_WIDTH(OUTPUT_WIDTH)
       )
   U_MEMORY_MAP (.*);
   
   bit_diff
     #(
       .WIDTH(INPUT_WIDTH)
       )
   U_BIT_DIFF (.*);
   
   // We aren't using the DMA for the AFU.
   assign dma.rd_addr = '0;
   assign dma.wr_addr = '0;
   assign dma.rd_size = '0;
   assign dma.wr_size = '0;
   assign dma.rd_go = 1'b0;
   assign dma.wr_go = 1'b0;
   assign dma.rd_en = 1'b0;
   assign dma.wr_en = 1'b0;
   assign dma.wr_data = '0;
            
endmodule




