// Copyright (c) 2023 University of Florida
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

// Greg Stitt
// University of Florida

// Module Name:  memory_map.sv
// Description:  This module implements a memory map for the bit_diff_module.   
//
//               Addresses follow all rules for CCI-P, which requires
//               even addresses for 64-bit data.
//
//               The memory map provides 2 inputs to the circuit:
//               go   : h0050,
//               data : h0052,
//
//               and provides two outputs to software:
//               result : h0054
//               done   : h0056
//
//==========================================================================

//==========================================================================
// Interface Description (All control signals are active high)
// clk : clk
// rst : rst (asynchronous)
// mmio : the mmio_if interface (see mmio_if.vh or afu.sv for explanation)
// go   : starts the DMA transfer
// data : the input to the bit_diff module
// result : the output from the bit_diff module
// done   : asserted when the bit_diff module is complete
//==========================================================================

module memory_map
  #(
    parameter int INPUT_WIDTH,
    parameter int OUTPUT_WIDTH
    )
   (
   input                          clk,
   input                          rst, 
                                  
                                  mmio_if.user mmio,
    
   output logic                   go,
   input logic                    done,
   output logic [INPUT_WIDTH-1:0] data,
   input logic [OUTPUT_WIDTH-1:0] result    
    );
   
   // =============================================================//   
   // MMIO write code
   // =============================================================//     
   always_ff @(posedge clk or posedge rst) begin 
      if (rst) begin
         go  <= '0;
         data  <= '0;
      end
      else begin
         if (mmio.wr_en == 1'b1) begin
            case (mmio.wr_addr)
              16'h0050: go   <= mmio.wr_data[0];
              16'h0052: data <= mmio.wr_data[$size(data)-1:0];
            endcase
         end
      end
   end

   // ============================================================= 
   // MMIO read code
   // =============================================================         
   always_ff @(posedge clk or posedge rst) begin
      if (rst) begin
         mmio.rd_data <= '0;
      end
      else begin             
         if (mmio.rd_en == 1'b1) begin
            
            mmio.rd_data <= '0;
            
            case (mmio.rd_addr)
              16'h0050: mmio.rd_data[0] <= go;
              16'h0052: mmio.rd_data[$size(data)-1:0] <= data;
              16'h0054: begin
                 // Sign-extend the result if negative.
                 if (result[$size(result)-1] == 1'b1)
                   mmio.rd_data <= '1;
                 
                 mmio.rd_data[$size(result)-1:0] <= result;

              end
              16'h0056: mmio.rd_data[0] <= done;
              
              // If the processor requests an address that is unused, return 0.
              default:  mmio.rd_data <= 64'h0;
            endcase
         end
      end
   end
endmodule
