// Copyright (c) 2023 University of Florida
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

// Greg Stitt
// University of Florida
//
// Description: This application demonstrates an AFU where the FPGA provides
// a bit difference calculator from my SystemVerilog tutorial.

#include <cstdlib>
#include <iostream>
#include <cmath>

#include <opae/utils.h>

#include "AFU.h"
// Contains application-specific information
#include "config.h"
// Auto-generated by OPAE's afu_json_mgr script
#include "afu_json_info.h"
#include "Timer.h"

using namespace std;

const unsigned NUM_TESTS = 100;
const int INPUT_WIDTH = 64;
const int OUTPUT_WIDTH = ceil(log2(2*INPUT_WIDTH+1));
const int TIMEOUT_SECONDS = 5;

// Software reference model
int bit_diff(unsigned long data) {

  int count = 0;

  for (unsigned i=0; i < 64; i++) {
    
    if (data & 0x01) count ++;
    else count --;

    data = data >> 1;                  
  }

  return count;
}


int main(int argc, char *argv[]) {

  unsigned errors = 0;

  try {
    // Create an AFU object to provide basic services for the FPGA. The 
    // constructor searches available FPGAs for one with an AFU with the
    // the specified ID.
    AFU afu(AFU_ACCEL_UUID); 

    cout << "Measured AFU Clock Frequency: " << afu.measureClock() / 1e6
         << "MHz" << endl;

    for (unsigned test=0; test < NUM_TESTS; test++) {

      cout << "Starting Test " << test << "...";

      // Send an input to the FPGA over MMIO.
      long data = rand();  
      afu.write(MMIO_DATA, (uint64_t) data);
    
      // Start the FPGA transfer.
      afu.write(MMIO_GO, 1);  

      // Wait until the FPGA is done.
      Timer wait_time;
      wait_time.start();
      while (afu.read(MMIO_DONE) == 0) {

        // Quit the program if the hardware doesn't assert done quickly enough.
        if (wait_time.elapsedTime() > TIMEOUT_SECONDS)
          throw TimeoutException();
        
        // Define this macro to slow down software requests for the done signal
        // during simulation.
#ifdef SLEEP_WHILE_WAITING
        this_thread::sleep_for(chrono::milliseconds(SLEEP_MS));
#endif
      }
      wait_time.stop();
        
      long result = afu.read(MMIO_RESULT);
      int correct_result = bit_diff(data);
      
      if (result != correct_result) {
        cout << "ERROR: hardware result " << result << " instead of " << correct_result << endl;
        errors ++;
      }      
      else {
        cout << "SUCCEEDED" << endl;
      }
    } 

    cout << "Total Errors: " << errors << endl;
    return EXIT_SUCCESS;
  }

  // Exception handling for all the runtime errors that can occur within 
  // the AFU wrapper class.
  catch (const fpga_result& e) {    
    
    // Provide more meaningful error messages for each exception.
    if (e == FPGA_BUSY) {
      cerr << "ERROR: All FPGAs busy." << endl;
    }
    else if (e == FPGA_NOT_FOUND) { 
      cerr << "ERROR: FPGA with accelerator " << AFU_ACCEL_UUID 
           << " not found." << endl;
    }
    else {
      // Print the default error string for the remaining fpga_result types.
      cerr << "ERROR: " << fpgaErrStr(e) << endl;    
    }
  }
  catch (const runtime_error& e) {    
    cerr << e.what() << endl;
  }
  catch (const opae::fpga::types::no_driver& e) {
    cerr << "ERROR: No FPGA driver found." << endl;
  }
  catch (const TimeoutException& e) {
    cerr << e.what() << endl;
  }

  return EXIT_FAILURE;
}


